# -*- coding: utf-8 -*-
# This file is part of the investment-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

import uuid
from datetime import timedelta
import matplotlib
import matplotlib.pyplot as plt
from decimal import Decimal
from io import BytesIO
from trytond.model import ModelSQL, fields, Unique
from trytond.pool import Pool
from trytond.exceptions import UserError
from trytond.i18n import gettext
from trytond.transaction import Transaction


IMG_PIXEL_WIDTH = 90
IMG_PIXEL_HEIGHT = 30
LINE_COLOR = '#aa0000'
BACKGROUND_COLOR = 'white'


class Chart(ModelSQL):
    'Chart Image'
    __name__ = 'investment_chart.img'

    asset = fields.Many2One(
        string='Asset', required=True,
        model_name='investment.asset', ondelete='CASCADE')
    imgdata = fields.Binary(string='Chart', required=True)
    fileuuid = fields.Char(string='UUID', required=True)

    @classmethod
    def __setup__(cls):
        super(Chart, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('uuid_uniq',
                Unique(t, t.fileuuid),
                'investment_chart.msg_uuid_unique'),
            ]

    @classmethod
    def _generate_uuid(cls):
        """ default: uuid
        """
        return str(uuid.uuid4())

    @classmethod
    def _check_uuid(cls, uuid_to_check):
        """ test if string is valid uuid
        """
        try:
            if isinstance(uuid_to_check, uuid.UUID):
                return
            else:
                uuid.UUID(uuid_to_check, version=4)
            return
        except Exception:
            raise UserError(gettext(
                'investment_chart.msg_invalid_uuid',
                uuidtxt=str(uuid_to_check)))

    @classmethod
    def validate(cls, charts):
        """ check uuid
        """
        super(Chart, cls).validate(charts)
        for chart in charts:
            cls._check_uuid(chart.fileuuid)

    @classmethod
    def prepare_data(cls, asset):
        """ read data from asset, build x/y-data
        """
        pool = Pool()
        Rates = Pool().get('investment.rate')
        IrDate = pool.get('ir.date')
        context = Transaction().context

        plot_date = context.get('pdate', IrDate.today())
        plot_days = context.get('pdays', 365)

        end_date = plot_date
        start_date = end_date - timedelta(days=plot_days)

        rates = Rates.search([
            ('asset', '=', asset.id),
            ('date', '>=', start_date),
            ('date', '<=', end_date),
            ], order=[('date', 'ASC')])

        x_data = []
        y_data = []
        min_point = Decimal('0.0')
        if len(rates) == 0:
            x_data = [start_date, end_date]
            y_data = [min_point, min_point]
        else:
            x_data = [x.date for x in rates]
            y_data = [x.rate for x in rates]
            min_point = min(y_data)

            # fix date ranges
            # start
            if x_data[0] > start_date:
                x_data.insert(0, start_date)
                y_data.insert(0, min_point)
            # end
            if x_data[-1] < end_date:
                x_data.append(end_date)
                y_data.append(min_point)
        return (x_data, y_data, min_point)

    @classmethod
    def get_plot_data(cls, assets):
        """ generate png-images from rates
        """
        result = []
        for asset in assets:
            (x_data, y_data, min_point) = cls.prepare_data(asset)

            # prepare plot area
            # resulting image has a 4 pixels width frame, we prepare
            # the plot area bigger, to be able to cut the frame
            plt.clf()
            fig = plt.figure(
                figsize=(
                    (IMG_PIXEL_WIDTH + 6) / 100,
                    (IMG_PIXEL_HEIGHT + 6) / 100.0),
                frameon=True,
                dpi=200.0,
                layout='compressed',
                facecolor=BACKGROUND_COLOR)
            ax = fig.add_subplot()
            ax.axis('off')
            ax.margins(0)

            # plot shaded area below line
            ax.fill_between(
                x_data, [min_point for x in x_data], y_data,
                alpha=0.15, color=LINE_COLOR, linewidth=0)
            # plot line
            ax.plot(
                x_data, y_data,
                linestyle='-', linewidth=0.5, color=LINE_COLOR,
                antialiased=True)

            with BytesIO() as fhdl:
                plt.savefig(
                    fname=fhdl,
                    format='png',
                    transparent=False,
                    dpi=100.0,
                    #           (xmin, ymin, xmax, ymax)
                    bbox_inches=matplotlib.transforms.Bbox(
                        [[0.03, 0.03],
                            [(IMG_PIXEL_WIDTH + 3) / 100.0,
                                (IMG_PIXEL_HEIGHT + 3) / 100.0]]),
                    pil_kwargs={'optimize': True, 'format': 'PNG'})
                result.append(fhdl.getvalue())
            plt.close()

        return result

    @classmethod
    def update_chart_images(cls, assets):
        """ update/create images for assets
        """
        pool = Pool()
        Chart2 = pool.get('investment_chart.img')
        Asset = pool.get('investment.asset')

        to_delete = [y for x in assets for y in x.chart_reference]
        if len(to_delete) > 0:
            Chart2.delete(to_delete)

        to_write = []
        for asset in assets:
            to_write.extend([
                [asset],
                {
                    'chart_reference': [('create', [{}])],
                }])
        if len(to_write) > 0:
            Asset.write(*to_write)

    @classmethod
    def create(cls, vlist):
        """ add uuid
        """
        pool = Pool()
        Asset = pool.get('investment.asset')

        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if 'fileuuid' not in values.keys():
                values['fileuuid'] = cls._generate_uuid()
            if 'imgdata' not in values.keys():
                values['imgdata'] = cls.get_plot_data([
                        Asset(values['asset']),
                    ])[0]
        return super(Chart, cls).create(vlist)

# end Chart
