# -*- coding: utf-8 -*-
# This file is part of the investment-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from datetime import date
from decimal import Decimal
from trytond.tests.test_tryton import with_transaction
from trytond.tests.test_tryton import activate_module, drop_db
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.exceptions import UserError
from trytond.modules.investment.tests.test_module import InvestmentTestCase
from .rates import rates_list


WRITE_DEBUG_FILES = False
IMG_BYTES_NUM = 1932


class ChartTestCase(InvestmentTestCase):
    'Test chart module'
    module = 'investment_chart'

    @classmethod
    def setUpClass(cls):
        drop_db()
        activate_module([
            'investment_chart', 'investment_watchlist'], 'en')

    def prep_chart_create(self):
        """ create asset
        """
        Asset = Pool().get('investment.asset')

        company = self.prep_asset_company()
        with Transaction().set_context({'company': company.id}):
            asset = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 1'))
            self.assertEqual(asset.symbol, 'usd/u')

            # add rates to asset
            Asset.write(*[
                [asset],
                {
                    'rates': [('create', rates_list)],
                }])
            self.assertEqual(len(asset.rates), 240)
        return (company, asset)

    @with_transaction()
    def test_chart_create(self):
        """ create asset
        """
        pool = Pool()
        Chart = pool.get('investment_chart.img')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):
            chart, = Chart.create([{
                'asset': asset.id,
                'imgdata': b'123',
                }])
            self.assertEqual(
                chart.asset.rec_name,
                'Product 1 | 21.9800 usd/u | 12/06/2022')
            self.assertEqual(len(chart.fileuuid), 36)
            self.assertEqual(asset.chart_img, b'123')

            # check deny of invalid uuid
            self.assertRaisesRegex(
                UserError,
                "The UUID 'not-a-uuid' is invalid.",
                Chart.write,
                *[
                    [chart],
                    {
                        'fileuuid': 'not-a-uuid',
                    }
                ])

    @with_transaction()
    def test_chart_check_duplicate_uuid(self):
        """ create asset, check deny of duplicate uuid
        """
        pool = Pool()
        Chart = pool.get('investment_chart.img')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):
            chart, = Chart.create([{
                'asset': asset.id,
                'imgdata': b'123',
                }])
            self.assertEqual(
                chart.asset.rec_name,
                'Product 1 | 21.9800 usd/u | 12/06/2022')
            self.assertEqual(len(chart.fileuuid), 36)

            # deny duplicate uuid
            asset2 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 2'))

            chart2, = Chart.create([{
                    'asset': asset2.id,
                    'imgdata': b'234',
                }])

            self.assertRaisesRegex(
                UserError,
                "This UUID already exist.",
                Chart.write,
                *[
                    [chart2],
                    {'fileuuid': chart.fileuuid},
                ])

    @with_transaction()
    def test_chart_check_cascade_delete(self):
        """ create asset, check delete of chart
        """
        pool = Pool()
        Chart = pool.get('investment_chart.img')
        Asset = pool.get('investment.asset')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):
            chart, = Chart.create([{
                'asset': asset.id,
                'imgdata': b'123',
                }])
            self.assertEqual(
                chart.asset.rec_name,
                'Product 1 | 21.9800 usd/u | 12/06/2022')

            Asset.delete([asset])
            self.assertEqual(len(Chart.search([])), 0)

    @with_transaction()
    def test_chart_plot_chart(self):
        """ create asset, plot chart, store to db
        """
        pool = Pool()
        Chart = pool.get('investment_chart.img')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({
                'company': company.id,
                'pdate': date(2022, 10, 30),
                'pdays': 180}):

            chart, = Chart.create([{'asset': asset.id}])
            self.assertEqual(
                chart.asset.rec_name,
                'Product 1 | 21.9800 usd/u | 12/06/2022')
            self.assertEqual(len(chart.asset.rates), 240)

            (x_data, y_data, min_val) = Chart.prepare_data(chart.asset)
            self.assertEqual(len(x_data), 130)  # not each date has a value
            self.assertEqual(len(y_data), 130)
            # 2022-10-30 -180 = 2022-05-03
            self.assertEqual(x_data[0], date(2022, 5, 3))
            self.assertEqual(x_data[-1], date(2022, 10, 30))
            self.assertEqual(y_data[0], Decimal('29.44'))
            self.assertEqual(y_data[-1], min_val)

            # write to file
            if WRITE_DEBUG_FILES:
                with open('investment.img.png', 'wb') as fhdl:
                    fhdl.write(chart.imgdata)

            with Transaction().set_context({
                    'investment_chart.img.imgdata': 'size'}):
                chart2, = Chart.browse([chart])
                self.assertEqual(chart2.imgdata, 1932)

    @with_transaction()
    def test_chart_check_plot_at_watchlist_entry(self):
        """ create asset, read imgdata from watchlist-entry
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        Chart = pool.get('investment_chart.img')
        Watchlist = pool.get('investment_watchlist.list')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):
            watchlist, = Watchlist.create([{
                'name': 'Watch List',
                'currency': company.currency.id,
                'company': company.id,
                'price_digits': 2,
                'uom_digits': 2,
                'lines': [('create', [{
                    'asset': asset.id,
                    'date': date(2022, 5, 3),
                    'price': Decimal('22.0'),
                    'quantity': Decimal('1.0'),
                    'expenses': Decimal('0.5'),
                    }])],
                }])

            with Transaction().set_context({
                    'pdate': date(2022, 10, 30),
                    'pdays': 180}):
                Chart.update_chart_images([asset])

            with Transaction().set_context({
                    'investment_chart.img.imgdata': 'size'}):
                assets = Asset.browse([asset])
                self.assertEqual(len(assets), 1)
                self.assertEqual(assets[0].chart_img, IMG_BYTES_NUM)

                watchlist2, = Watchlist.browse([watchlist])
                self.assertEqual(len(watchlist2.lines), 1)
                self.assertEqual(
                    watchlist2.lines[0].rec_name,
                    'Product 1 | 21.98 usd (-2.31 %, -0.52 usd) | 12/06/2022')
                self.assertEqual(watchlist2.lines[0].chart_img, IMG_BYTES_NUM)

    @with_transaction()
    def test_chart_plot_empty_chart(self):
        """ create asset, plot chart without data, store to db
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        Chart = pool.get('investment_chart.img')

        (company, asset) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):

            # delete rates
            Asset.write(*[
                [asset],
                {
                    'rates': [('delete', [x.id for x in asset.rates])],
                }])
            self.assertEqual(len(asset.rates), 0)

            chart, = Chart.create([{
                'asset': asset.id,
                }])
            self.assertEqual(chart.asset.rec_name, 'Product 1 | - usd/u | -')
            self.assertEqual(len(chart.asset.rates), 0)

            with Transaction().set_context({
                    'pdate': date(2022, 10, 30),
                    'pdays': 180}):
                (x_data, y_data, min_val) = Chart.prepare_data(chart.asset)
                self.assertEqual(len(x_data), 2)
                self.assertEqual(len(y_data), 2)
                self.assertEqual(x_data[0], date(2022, 5, 3))
                self.assertEqual(x_data[-1], date(2022, 10, 30))
                self.assertEqual(y_data[0], Decimal('0.0'))
                self.assertEqual(y_data[-1], Decimal('0.0'))

            # write to file
            if WRITE_DEBUG_FILES:
                with open('investment.emtpy.png', 'wb') as fhdl:
                    fhdl.write(chart.imgdata)

            with Transaction().set_context({
                    'investment_chart.img.imgdata': 'size'}):
                chart2, = Chart.browse([chart])
                self.assertEqual(chart2.imgdata, 241)

    @with_transaction()
    def test_chart_plot_multiple_chart(self):
        """ create asset, plot multiple chart
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        Chart = pool.get('investment_chart.img')

        (company, asset1) = self.prep_chart_create()
        (company, asset2) = self.prep_chart_create()
        (company, asset3) = self.prep_chart_create()
        with Transaction().set_context({'company': company.id}):
            asset1.product.template.name = 'Prod 1'
            asset1.product.template.save()
            asset2.product.template.name = 'Prod 2'
            asset2.product.template.save()
            asset3.product.template.name = 'Prod 3'
            asset3.product.template.save()

            assets = Asset.browse([asset1, asset2, asset3])
            self.assertEqual(len(assets), 3)
            self.assertEqual(
                assets[0].rec_name,
                'Prod 1 | 21.9800 usd/u | 12/06/2022')
            self.assertEqual(
                assets[1].rec_name,
                'Prod 2 | 21.9800 usd/u | 12/06/2022')
            self.assertEqual(
                assets[2].rec_name,
                'Prod 3 | 21.9800 usd/u | 12/06/2022')

            with Transaction().set_context({
                    'pdate': date(2022, 10, 30),
                    'pdays': 180}):
                Chart.update_chart_images(assets)

            assets = Asset.browse([asset1, asset2, asset3])
            self.assertEqual(len(assets), 3)

            if WRITE_DEBUG_FILES:
                with open('chart_img-1.png', 'wb') as fhdl:
                    fhdl.write(assets[0].chart_img)
                with open('chart_img-2.png', 'wb') as fhdl:
                    fhdl.write(assets[1].chart_img)
                with open('chart_img-3.png', 'wb') as fhdl:
                    fhdl.write(assets[2].chart_img)

            with Transaction().set_context({
                    'investment_chart.img.imgdata': 'size'}):
                assets = Asset.browse([asset1, asset2, asset3])
                self.assertEqual(len(assets), 3)
                self.assertEqual(assets[0].chart_img, IMG_BYTES_NUM)
                self.assertEqual(assets[1].chart_img, IMG_BYTES_NUM)
                self.assertEqual(assets[2].chart_img, IMG_BYTES_NUM)

# end ChartTestCase


del InvestmentTestCase
